import logging

from pymchelper.estimator import Estimator

logger = logging.getLogger(__name__)


class ExcelWriter:
    """
    Supports writing XLS files (MS Excel 2003 format)
    """

    def __init__(self, filename, options):
        self.filename = filename
        if not self.filename.endswith(".xls"):
            self.filename += ".xls"

    def write(self, estimator: Estimator):
        try:
            import xlwt
        except ImportError as e:
            logger.error("Generating Excel files not available on your platform (please install xlwt).")
            raise e

        # create workbook
        wb = xlwt.Workbook()

        for page_id, page in enumerate(estimator.pages):

            # save only 1-D data
            if page.dimension != 1:
                logger.warning("page dimension {:d} != 1, XLS output not supported".format(estimator.dimension))
                return 1

            # create worksheet
            ws = wb.add_sheet(f'Data_p{page_id}')

            # save X axis data
            for i, x in enumerate(page.plot_axis(0).data):
                ws.write(i, 0, x)

            # save Y axis data
            for i, y in enumerate(page.data_raw):
                ws.write(i, 1, y)

            # save error column (if present)
            if page.error_raw is not None:
                for i, e in enumerate(page.error_raw):
                    ws.write(i, 2, e)

        # save file
        logger.info("Writing: " + self.filename)
        wb.save(self.filename)

        return 0



from enum import IntEnum
import logging
import gc
import os
from collections import defaultdict
from glob import glob
from pathlib import Path
from typing import List, Optional

from pymchelper.averaging import (Aggregator, SumAggregator, WeightedStatsAggregator, ConcatenatingAggregator,
                                  NoAggregator)
from pymchelper.estimator import ErrorEstimate, Estimator, average_with_nan
from pymchelper.readers.topas import TopasReaderFactory
from pymchelper.readers.fluka import FlukaReader, FlukaReaderFactory
from pymchelper.readers.shieldhit.general import SHReaderFactory
from pymchelper.readers.shieldhit.reader_base import SHReader
from pymchelper.writers.common import Converters

logger = logging.getLogger(__name__)


class AggregationType(IntEnum):
    """
    Enum for different types of aggregation.
    This enum is related to integer value stored in SHIELD-HIT12A binary files,
    which defines how the data is aggregated.
    Below few examples of how such aggregation is used in SHIELD-HIT12A:
      - NoAggregation is used for density (RHO) and material scorer.
      - Sum is used for particle counter (COUNT).
      - AveragingCumulative is used for dose and fluence scorers.
      - AveragingPerPrimary is used for LET scorers (TLET and DLET).
      - Concatenation is used for phase space (MCPL) scorer.
    """

    NoAggregation = 0
    Sum = 1
    AveragingCumulative = 2
    AveragingPerPrimary = 3
    Concatenation = 4


def guess_reader(filename):
    """
    Guess a reader based on file contents or extensions.
    In some cases (i.e. binary SHIELD-HIT12A files) access to file contents is needed.
    :param filename:
    :return: Instantiated reader object
    """
    reader = None
    fluka_reader = FlukaReaderFactory(filename).get_reader()
    if fluka_reader:
        reader = fluka_reader(filename)
    else:
        sh_reader = SHReaderFactory(filename).get_reader()
        if sh_reader:
            reader = sh_reader(filename)
        else:
            topas_reader = TopasReaderFactory(filename).get_reader()
            if topas_reader:
                reader = topas_reader(filename)
    return reader


def guess_corename(filename):
    """
    Guess a reader based on file contents or extensions.
    In some cases (i.e. binary SHIELD-HIT12A files) access to file contents is needed.
    :param filename:
    :return: the corename of the file (i.e. the basename without the running number for averaging)
    """
    corename = FlukaReader(filename).corename
    if corename is None:
        corename = SHReader(filename).corename
    return corename


def fromfile(filename: str) -> Optional[Estimator]:
    """
    Read estimator data from a binary file `filename`
    Note that for the in some cases the data are post-processes (i.e. normalized) after reading.
    For example SHIELD-HIT12A saves dose and fluence data as cumulative values,
    which are normalized by the number of primaries after by the Reader responsible for parsing binary files.
    This way dose and fluence (and other similar quantities) are saved in Estimator as "per primary" values.
    Fluka on the other hand saves dose and fluence as "per primary" values, so no normalization is needed.
    """

    reader = guess_reader(filename)
    if reader is None:
        raise Exception("File format not compatible", filename)
    estimator = Estimator()
    estimator.file_counter = 1
    if not reader.read(estimator):  # some problems occurred during read
        logger.error("Error reading file %s", filename)
        estimator = None
    return estimator


def fromfilelist(input_file_list,
                 error: ErrorEstimate = ErrorEstimate.stderr,
                 nan: bool = False) -> Optional[Estimator]:
    """
    Reads all files from a given list using `fromfile` method, and returns a list of averaged estimators.

    :param input_file_list: list of files to be read
    :param error: error estimation, see class ErrorEstimate class in pymchelper.estimator
    :param nan: if True, NaN (not a number) are excluded when averaging data.
    :return: list of estimators
    """
    if not isinstance(input_file_list, list):  # probably a string instead of list
        input_file_list = [input_file_list]

    if nan:
        estimator_list = [fromfile(filename) for filename in input_file_list]
        result = average_with_nan(estimator_list, error)
    elif len(input_file_list) == 1:
        result = fromfile(input_file_list[0])
        if not result:
            return None
    else:
        result = fromfile(input_file_list[0])
        if not result:
            return None

        # _aggregator_mapping maps SHIELD-HIT12A normalization types (integers) to pymchelper aggregators
        # using enums for clarity. AveragingCumulative (e.g., dose) and AveragingPerPrimary (e.g., LET)
        # both utilize WeightedStatsAggregator. SHIELD-HIT12A stores "cumulative-like" data (e.g., dose,
        # fluence) in BDO format as quantities for all particles. pymchelper normalizes this upon reading
        # a BDO file by the number of primaries, making the `estimator` object data pre-normalized. Hence,
        # aggregation for "cumulative-like" and "per-primary" data is handled uniformly in this mapping.
        _aggregator_mapping: dict[AggregationType, Aggregator] = {
            AggregationType.NoAggregation: NoAggregator,
            AggregationType.Sum: SumAggregator,
            AggregationType.AveragingCumulative: WeightedStatsAggregator,
            AggregationType.AveragingPerPrimary: WeightedStatsAggregator,
            AggregationType.Concatenation: ConcatenatingAggregator
        }

        # create aggregators for each page and fill them with data from first file
        page_aggregators = []
        for page in result.pages:

            # if no normalization attribute present (Fluka?) we can assume it is a cumulative-like quantity
            current_page_normalisation = getattr(page, 'page_normalized', AggregationType.AveragingCumulative.value)

            # guess the aggregator based on the normalisation type
            aggregator = _aggregator_mapping.get(current_page_normalisation, WeightedStatsAggregator)()
            logger.debug("Selected aggregator %s for page %s", aggregator, page.name)

            # feed the aggregator with data from the first file
            aggregator.update(value=page.data_raw, weight=result.number_of_primaries)
            page_aggregators.append(aggregator)

        # process all other files, if there are any
        for filename in input_file_list[1:]:
            current_estimator = fromfile(filename)
            for current_page, aggregator in zip(current_estimator.pages, page_aggregators):
                aggregator.update(value=current_page.data_raw, weight=current_estimator.number_of_primaries)

            # force garbage collection if the estimator is too large
            estimator_size_mbytes = sum(page.data_raw.nbytes for page in current_estimator.pages) / 1024 / 1024
            gc_threshold_mbytes = 100
            if estimator_size_mbytes > gc_threshold_mbytes:
                logger.info("Large estimator (%.1f MB) detected, performing garbage collection", estimator_size_mbytes)
                gc.collect()
            result.number_of_primaries += current_estimator.number_of_primaries

        # extract data from aggregators and fill then into the result
        for page, aggregator in zip(result.pages, page_aggregators):
            logger.debug("Extracting data from aggregator %s for page %s", aggregator, page.name)
            page.data_raw = aggregator.data
            page.error_raw = aggregator.error(error_type=error.name)

    result.file_counter = len(input_file_list)
    core_names_dict = group_input_files(input_file_list)
    if len(core_names_dict) == 1:
        result.file_corename = list(core_names_dict)[0]

    return result


def frompattern(pattern: str, error: ErrorEstimate = ErrorEstimate.stderr, nan: bool = True):
    """
    Reads all files matching pattern, e.g.: 'foobar_*.bdo', and returns a list of averaged estimators.

    :param pattern: pattern to be matched for reading.
    :param error: error estimation, see class ErrorEstimate class in pymchelper.estimator
    :param nan: if True, NaN (not a number) are excluded when averaging data.
    :return: a list of estimators, or an empty list if no files were found.
    """

    try:
        list_of_matching_files = glob(pattern)
    except TypeError as e:  # noqa: F841
        list_of_matching_files = pattern

    core_names_dict = group_input_files(list_of_matching_files)

    result = [fromfilelist(filelist, error, nan) for _, filelist in core_names_dict.items()]

    return result


def get_topas_estimators(output_files_path: str) -> List[Estimator]:
    """Get Topas estimators from provided directory"""
    estimators_list = []
    for path in Path(output_files_path).iterdir():
        topas_reader = TopasReaderFactory(str(path)).get_reader()
        if topas_reader:
            reader = topas_reader(path)
            estimator = Estimator()
            reader.read(estimator)
            estimators_list.append(estimator)

    return estimators_list


def convertfromlist(filelist, error, nan, outputdir, converter_name, options, outputfile=None):
    """
    :param filelist:
    :param error: error estimation, see class ErrorEstimate class in pymchelper.estimator
    :param nan: if True, NaN (not a number) are excluded when averaging data.
    :param outputdir:
    :param converter_name:
    :param options:
    :param outputfile:
    :return:
    """
    estimator = fromfilelist(filelist, error, nan)
    if not estimator:
        return None
    if outputfile is not None:
        output_path = outputfile
    elif outputdir is None:
        output_path = estimator.file_corename
    else:
        output_path = os.path.join(outputdir, estimator.file_corename)
    status = tofile(estimator, output_path, converter_name, options)
    return status


def convertfrompattern(pattern, outputdir, converter_name, options, error=ErrorEstimate.stderr, nan: bool = True):
    """

    :param pattern:
    :param outputdir:
    :param converter_name:
    :param options:
    :param error: error estimation, see class ErrorEstimate class in pymchelper.estimator
    :param nan: if True, NaN (not a number) are excluded when averaging data.
    :return:
    """
    list_of_matching_files = glob(pattern)

    core_names_dict = group_input_files(list_of_matching_files)

    status = []
    for _, filelist in core_names_dict.items():
        status.append(convertfromlist(filelist, error, nan, outputdir, converter_name, options))
    return max(status)


def tofile(estimator, filename, converter_name, options):
    """
    Save a estimator data to a ``filename`` using converter defined by ``converter_name``
    :param estimator:
    :param filename:
    :param converter_name:
    :param options:
    :return:
    """
    writer_cls = Converters.fromname(converter_name)
    writer = writer_cls(filename, options)
    logger.debug(f"File corename : {filename}")
    status = writer.write(estimator)
    return status


def group_input_files(input_file_list):
    """
    Takes set of input file names, belonging to possibly different estimators.
    Input files are grouped according to the estimators and for each group
    merging is performed, as in @merge_list method.
    Output file name is automatically generated.
    :param input_file_list: list of input files
    :return: core_names_dict
    """
    core_names_dict = defaultdict(list)
    # keys - core_name, value - list of full paths to corresponding files

    # loop over input list of file paths
    for filepath in input_file_list:
        core_name = guess_corename(filepath)
        core_names_dict[core_name].append(filepath)

    return core_names_dict



"""
Output from the simulation running on multiple processess needs to be agregated.
The simplest way of doing so is to calculated the average of the data.
There are however more sophisticated cases: COUNT scorers needs to be summed up, not averaged.
Phase space data needs to be concatenated, not averaged.
Each of the parallel jobs can have different number of histories, so the weighted average needs to be calculated.
Moreover, in case averaging is employed, we can estimate spread of the data as standard deviation or standard error.

The binary output files from each job may be quite large (even ~GB for scoring in fine 3D mesh), to obtain good
statistics we sometimes parallelise the simulation of hundreds or thousands of jobs (when using HPC clusters).
In such case it's not feasible to load all the data into memory and calculate the average in one go, using standard
functions from numpy library. Instead, we need to calculate the average in an online manner,
i.e. by updating the state of respective aggregator object with each new binary output file read.

Such approach results in a significant reduction of memory usage and is more numerically stable.

This module contains several classes for aggregating data from multiple files:
- Aggregator: base class for all other aggregators
- WeightedStatsAggregator: for calculating weighted (using weights which not necessarily sums up 1) mean and variance
- ConcatenatingAggregator: for concatenating data
- SumAggregator: for calculating sum instead of variance
- NoAggregator: for cases when no aggregation is required

All aggregators have `data` and `error` property, which can be used to obtain the result of the aggregation.
The `data` property returns the result of the aggregation: mean, sum or concatenated array.
The `error` property returns the spread of data for WeightedStatsAggregator, and `None` for other aggregators.

The `update` method is used to update the state of the aggregator with new data from the file.

For details on how this method is applied to average binary output of the MC codes,
see `fromfilelist` method from `input_output.py` module.
"""

from dataclasses import dataclass, field
import logging
from typing import Union, Optional
import numpy as np
from numpy.typing import ArrayLike


@dataclass
class Aggregator:
    """
    Base class for all aggregators.
    The `data` property returns the result of the aggregation, needs to be implemented in derived classes.
    The `error` function returns the spread of data, can be implemented in derived classes. It's a function,
    not a property as different type of error can be calculated (standard deviation, standard error, etc.).
    Type of errors may be then passed in optional keyword arguments `**kwargs`.
    """

    data: Union[float, ArrayLike] = float('nan')
    _updated: bool = field(default=False, repr=False, init=False)

    def update(self, value: Union[float, ArrayLike], **kwargs):
        """Update the state of the aggregator with new data."""
        raise NotImplementedError(f"Update function not implemented for {self.__class__.__name__}")

    def error(self, **kwargs):
        """Default implementation of error function, returns None."""
        logging.debug("Error calculation not implemented for %s", self.__class__.__name__)

    @property
    def updated(self) -> bool:
        """
        Check if the aggregator was updated. The newly created aggregator is in the state of not being updated.
        That means that no aggregation results are present via `data` or `error` properties.
        We rely on the fact that `data` attribute is set to `nan` on creation of the aggregator object.
        On first call to the `update` method the `data` is being filled with floating point numbers or arrays.
        The `update` method sets also `self._updated` to True.
        """
        if isinstance(self.data, float) and np.isnan(self.data):
            return False
        return self._updated


@dataclass
class WeightedStatsAggregator(Aggregator):
    """
    Calculates weighted mean and variance of a sequence of numbers or numpy arrays.
    We do not use frequency weights (which sums up to 1), the total sum of all weights is not known as
    we are aggregating data from multiple files with different number of histories.
    The aggregation uses single pass loop over all files.

    Good overview of currently known methods to calculate online weighted mean and variance can be found in [2].
    The original method to calculate online mean and variance was proposed by Welford in [1].
    The weighed version of this algoritm is nicely illustrated in [3].
    Here we employ algoritm proposed by West in [4] and descibed in Wikipedia [5].

    [1]  Welford, B. P. (1962). "Note on a method for calculating corrected sums of squares and products".
    Technometrics. 4 (3): 419–420.
    [2] Schubert, Erich, and Michael Gertz. "Numerically stable parallel computation of (co-) variance."
    Proceedings of the 30th international conference on scientific and statistical database management. 2018.
    [3] https://justinwillmert.com/posts/2022/notes-on-calculating-online-statistics/
    [4] West, D. H. D. (1979). "Updating Mean and Variance Estimates: An Improved Method".
    Communications of the ACM. 22 (9): 532-535.
    [5] https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Weighted_incremental_algorithm
    """

    data: Union[float, ArrayLike] = float('nan')

    _accumulator_S: Union[float, ArrayLike] = field(default=float('nan'), repr=False, init=False)
    _total_weight_squared: float = field(default=0., repr=False, init=False)
    total_weight: float = 0

    def update(self, value: Union[float, ArrayLike], weight: float = 1.0, **kwargs):
        """
        Update the state of the aggregator with new data.
        Note that the weights are so called "reliability weights", not frequency weights.
        If unsure put here the number of histories from the file if you are aggregating data from multiple files.
        """
        if weight < 0:
            raise ValueError("Weight must be non-negative")

        # first pass initialization
        if not self.updated:
            self.data = value * 0
            self._accumulator_S = value * 0

        # W_n = W_{n-1} + w_n
        self.total_weight += weight
        self._total_weight_squared += weight**2

        mean_old = self.data
        # mu_n = (1 - w_n / W_n) * mu_{n-1} + (w_n / W_n) * x_n
        # or in other words:
        # mu_n - mu_{n-1} = (w_n / W_n) * (x_n - mu_{n-1})
        self.data += (weight / self.total_weight) * (value - mean_old)

        self._accumulator_S += weight * (value - self.data) * (value - mean_old)

        self._updated = True
        logging.debug("Updated aggregator with value %s and weight %s", value, weight)

    @property
    def mean(self) -> Union[float, ArrayLike]:
        """Weighted mean of the sample"""
        return self.data

    @property
    def variance_population(self) -> Union[float, ArrayLike]:
        """Biased estimate of the variance"""
        if not self.updated:
            raise ValueError("No data to calculate variance")
        if self.total_weight <= 0:
            raise ValueError("Total weight must be positive")
        return self._accumulator_S / self.total_weight

    @property
    def variance_sample(self) -> Union[float, ArrayLike]:
        """
        Unbiased estimate of the variance.
        The bias of the weighted estimator if (1 - sum w_i^2 / W_n^2), or in other words:
        1 - "sum of squares of weights" / "square of sum of weights".
        For all equal weights the bias is 1 - n * w^2/((n * w)^2) = 1 - 1/n which
        leads to the well known formula for the sample variance.
        Here we use the weighted version of the formula.
        """
        if not self.updated:
            raise ValueError("No data to calculate variance")
        if self.total_weight <= 0:
            raise ValueError("Total weight must be positive")
        return self._accumulator_S / (self.total_weight - (self._total_weight_squared / self.total_weight))

    @property
    def stddev(self) -> Union[float, ArrayLike]:
        """Standard deviation of the sample"""
        return np.sqrt(self.variance_sample)

    @property
    def stderr(self) -> Union[float, ArrayLike]:
        """
        Standard error of the sample.
        For weighted data it is calculated as:
        stddev * sqrt(sum w_i^2) / sum w_i
        For equal weights it reduces via sqrt( n * w^2) / (n * w) = stddev / sqrt(n)
        """
        return self.stddev * np.sqrt(self._total_weight_squared) / self.total_weight

    def error(self, **kwargs) -> Optional[Union[float, ArrayLike]]:
        """
        Error calculation function, can be used to calculate standard deviation or standard error.
        Type of error may be requested by `error_type` keyword argument with `stddev` or `stderr` values.
        For other values or if the keyword argument is not present, None is returned.
        """
        logging.debug("Calculating error with kwargs: %s", kwargs)
        if 'error_type' in kwargs:
            if kwargs['error_type'] == 'stddev':
                return self.stddev
            if kwargs['error_type'] == 'stderr':
                return self.stderr
        return None


@dataclass
class ConcatenatingAggregator(Aggregator):
    """Class for concatenating numpy arrays"""

    def update(self, value: Union[float, ArrayLike], **kwargs):
        """Update the state of the aggregator with new data."""
        if not self.updated:
            self.data = value
        else:
            self.data = np.concatenate((self.data, value))
        self._updated = True


@dataclass
class SumAggregator(Aggregator):
    """Class for calculating sum of a sequence of numbers."""

    def update(self, value: Union[float, ArrayLike], **kwargs):
        """Update the state of the aggregator with new data."""
        # first value added
        if not self.updated:
            self.data = value
        # subsequent values added
        else:
            self.data += value
        self._updated = True


@dataclass
class NoAggregator(Aggregator):
    """
    Class for cases when no aggregation is required.
    Sets the data to the first value and does not update it.
    """

    def update(self, value: Union[float, ArrayLike], **kwargs):
        """Update the state of the aggregator with new data."""
        # set value only on first update
        if not self.updated:
            logging.debug("Setting data to %s", value)
            self.data = value
        self._updated = True
